# rspec-power_assert
require 'timesteps'
require 'rspec-power_assert'

describe "Fixed360Day.new" do

  example 'full arguments' do

    is_asserted_by { 
      DateTime::Fixed360Day.new(-4712,1,1,0,0,0,0).is_a?(DateTime::Fixed360Day) 
    }

    d = DateTime::Fixed360Day.new(2001, 3, 5, 9, 35, 12.5, 9.quo(24))

    is_asserted_by { d.year   == 2001 }
    is_asserted_by { d.month  == 3    }
    is_asserted_by { d.day    == 5    }
    is_asserted_by { d.hour   == 9    }
    is_asserted_by { d.minute == 35   }
    is_asserted_by { d.second + d.second_fraction == 12.5 }
    is_asserted_by { d.offset == 9.quo(24) }

  end  

  example 'ommitted arguments' do

    is_asserted_by { 
      DateTime::Fixed360Day.new().is_a?(DateTime::Fixed360Day) 
    }

    d = DateTime::Fixed360Day.new()

    is_asserted_by { d.year   == -4712 }
    is_asserted_by { d.month  == 1     }
    is_asserted_by { d.day    == 1     }
    is_asserted_by { d.hour   == 0     }
    is_asserted_by { d.minute == 0     }
    is_asserted_by { d.second + d.second_fraction == 0.0   }
    is_asserted_by { d.offset == 0     }

  end  

  example "raised for invalid date" do

    # leap day for noleap calendar
    is_asserted_by { 
      DateTime::Fixed360Day.new(2020,2,29).is_a?(DateTime::Fixed360Day) 
    }

    # date not exist
    expect { DateTime::Fixed360Day.new(2020,1,31) }.to raise_error(RuntimeError)
    expect { DateTime::Fixed360Day.new(2020,3,31) }.to raise_error(RuntimeError)
    expect { DateTime::Fixed360Day.new(2020,5,31) }.to raise_error(RuntimeError)
    expect { DateTime::Fixed360Day.new(2020,7,31) }.to raise_error(RuntimeError)
    expect { DateTime::Fixed360Day.new(2020,8,31) }.to raise_error(RuntimeError)
    expect { DateTime::Fixed360Day.new(2020,10,31) }.to raise_error(RuntimeError)
    expect { DateTime::Fixed360Day.new(2020,12,31) }.to raise_error(RuntimeError)

    # date not exist
    expect { DateTime::Fixed360Day.new(2020,4,31) }.to raise_error(RuntimeError)

  end

end

describe "Fixed360Day.jd" do

  example "epoch" do

    d = DateTime::Fixed360Day.jd(0)    
    is_asserted_by { d.jd == 0 }

  end

  example "unixpoch" do

    jd = DateTime::Fixed360Day::UNIX_EPOCH_IN_AJD.ceil

    d = DateTime::Fixed360Day.jd(jd)    
    is_asserted_by { d.jd == jd }
    is_asserted_by { d.year  == 1970 }
    is_asserted_by { d.month == 1 }
    is_asserted_by { d.day   == 1 }

  end

end

describe "Fixed360Day#<=>" do

  example {

    d1 = DateTime::Fixed360Day.new(2001, 3, 5, 9, 35, 12.5, 9.quo(24))
    d2 = DateTime::Fixed360Day.new(2001, 3, 5, 9, 35, 12.5, 9.quo(24))

    is_asserted_by { d1 == d2 }
    
  }
      
end

describe "Fixed360Day.parse_timestamp" do

  example '2001-03-05T09:35:12.5+09:00' do

    d1 = DateTime::Fixed360Day.new(2001, 3, 5, 9, 35, 12.5, 9.quo(24))

    [
      '2001-03-05T09:35:12.5+09:00',
      '2001-03-05 09:35:12.5 +09:00',
      '2001-03-05 09:35:12.5 +0900',
      '2001-03-05 09:35:12.5 +900',
      '2001-03-05 09:35:12.5 +09',
      '2001-03-05 09:35:12.5 +9',
      '2001/03/05 09:35:12.5 +09:00',
      '2001-3-5 9:35:12.5 +09:00',
      '05,Mar,2001 09:35:12.5 +09:00',
      '05-Mar-2001 09:35:12.5 +09:00',
    ].each do |string|
      
      d2 = DateTime.parse_timestamp(string, calendar: "360_day")

      is_asserted_by { d2 == d1 }
    end

  end  

  example 'raise for "2000-02-29"' do

    # leap day for noleap calendar
    is_asserted_by { DateTime.parse_timestamp("2000-2-29", calendar: "360_day").is_a?(DateTime::Fixed360Day) }
    
    
  end

  example '2000-12-31 24:00:00' do

    d1 = DateTime::Fixed360Day.new(2001, 1, 1, 0, 0, 0, 0)
    d2 = DateTime.parse_timestamp('2000-12-30 24:00:00', calendar: "360_day")

    is_asserted_by { d2 == d1 }

  end
  

end

describe "Fixed360Day#jd" do

  example 'epoch' do

    d = DateTime::Fixed360Day.new()    
    is_asserted_by { d.jd == 0 }

  end

  example 'few-years' do

    d = DateTime::Fixed360Day.new(-4712,12,30)    
    is_asserted_by { d.jd == 359 }

    d = DateTime::Fixed360Day.new(-4711,1,1)    
    is_asserted_by { d.jd == 360 }

    d = DateTime::Fixed360Day.new(-4708,1,1)    
    is_asserted_by { d.jd == 360*4 }

    d = DateTime::Fixed360Day.new(-4713,1,1)    
    is_asserted_by { d.jd == -360 }

    d = DateTime::Fixed360Day.new(-4716,1,1)    
    is_asserted_by { d.jd == -360*4 }

  end

end

describe "Fixed360Day#ajd" do

  example 'epoch' do

    d = DateTime::Fixed360Day.new()    
    is_asserted_by { d.ajd == -1.quo(2) }

  end

  example 'few-years' do

    d = DateTime::Fixed360Day.new(-4712,12,30)    
    is_asserted_by { d.ajd == 359-1.quo(2) }

    d = DateTime::Fixed360Day.new(-4711,1,1)    
    is_asserted_by { d.ajd == 360-1.quo(2) }

    d = DateTime::Fixed360Day.new(-4708,1,1)    
    is_asserted_by { d.ajd == 360*4-1.quo(2) }

    d = DateTime::Fixed360Day.new(-4713,1,1)    
    is_asserted_by { d.ajd == -360-1.quo(2) }

    d = DateTime::Fixed360Day.new(-4716,1,1)    
    is_asserted_by { d.ajd == -360*4-1.quo(2) }

  end

end

describe "Fixed360Day#strftime" do

  example 'standard' do

    d = DateTime::Fixed360Day.new(2001, 3, 5, 9, 35, 12.5, 9.quo(24))

    is_asserted_by { d.strftime("%FT%T.%L%:z")  == "2001-03-05T09:35:12.500+09:00" }
    is_asserted_by { d.strftime("%Y%m%d%H%M%S") == "20010305093512" }

  end

end


describe "Fixed360Day#next_XXX" do

  example "next_year" do 
    
    d = DateTime::Fixed360Day.new(2001, 3, 5, 9, 35, 12.5, 9.quo(24))
    
    is_asserted_by { d.next_year().year   == 2002 }
    is_asserted_by { d.next_year(10).year == 2011 }

    is_asserted_by { d.prev_year().year   == 2000 }
    is_asserted_by { d.prev_year(10).year == 1991 }
  
  end

  example "next_month" do 
    
    d = DateTime::Fixed360Day.new(2001, 3, 5, 9, 35, 12.5, 9.quo(24))
    
    is_asserted_by { d.next_month().month   == 4 }
    is_asserted_by { d.next_month(10).month == 1 }
    is_asserted_by { (d>>1).month           == 4 }
    is_asserted_by { (d>>10).month          == 1 }

    is_asserted_by { d.prev_month().month   == 2 }
    is_asserted_by { d.prev_month(10).month == 5 }
    is_asserted_by { (d<<1).month           == 2 }
    is_asserted_by { (d<<10).month          == 5 }
  
    is_asserted_by { (d>>12).year    == 2002 }
    is_asserted_by { (d>>10*12).year == 2011 }
    is_asserted_by { (d<<12).year    == 2000 }
    is_asserted_by { (d<<10*12).year == 1991 }
    
  end

  example "next_day" do 

    d = DateTime::Fixed360Day.new(2001, 3, 5, 9, 35, 12.5, 9.quo(24))
    
    is_asserted_by { d.next_day().day   == 6 }
    is_asserted_by { d.next_day(10).day == 15 }
    is_asserted_by { d.next_day(30).day == 5 }

    is_asserted_by { (d + 1).day  == 6 }
    is_asserted_by { (d + 10).day == 15 }
    is_asserted_by { (d + 30).day == 5 }

    is_asserted_by { d.prev_day().day   == 4 }
    is_asserted_by { d.prev_day(10).day == 25 }
    is_asserted_by { d.prev_day(30).day == 5 }

    is_asserted_by { (d - 1).day  == 4 }
    is_asserted_by { (d - 10).day == 25 }
    is_asserted_by { (d - 30).day == 5 }

  end

end

describe "Fixed360Day#fraction" do

  example do 
    
    d1 = DateTime::Fixed360Day.new(2001, 3, 5, 9, 35, 12.5, 9.quo(24))
    d2 = DateTime::Fixed360Day.new(2001, 3, 5, 9, 35, 12.893, 9.quo(24))
    
    is_asserted_by { d1.fraction == 9.quo(24) + 35.quo(1440) + 12.5.to_r.quo(86400) }
    is_asserted_by { d2.fraction == 9.quo(24) + 35.quo(1440) + 12.893.to_r.quo(86400) }
    
  end

end

describe "Fixed360Day#difference_in_years" do

  example do 
    
    d1 = DateTime::Fixed360Day.new(2001, 3, 5, 9, 35, 12.5, 9.quo(24))
    d2 = DateTime::Fixed360Day.new(2002, 3, 4, 9, 35, 12.5, 9.quo(24))
    d3 = DateTime::Fixed360Day.new(2002, 3, 5, 9, 35, 11.5, 9.quo(24))
    d4 = DateTime::Fixed360Day.new(2002, 3, 5, 9, 35, 12.5, 9.quo(24))
    
    is_asserted_by { d2.difference_in_years(d1) == 1.quo(2) }
    is_asserted_by { d3.difference_in_years(d1) == 1.quo(2) }
    is_asserted_by { d4.difference_in_years(d1) == 1 }
    
  end

end

describe "Fixed360Day#difference_in_months" do

  example do 
    
    d1 = DateTime::Fixed360Day.new(2001, 3, 5, 9, 35, 12.5, 9.quo(24))
    d2 = DateTime::Fixed360Day.new(2001, 4, 4, 9, 35, 12.5, 9.quo(24))
    d3 = DateTime::Fixed360Day.new(2001, 4, 5, 9, 35, 11.5, 9.quo(24))
    d4 = DateTime::Fixed360Day.new(2001, 4, 5, 9, 35, 12.5, 9.quo(24))
    d5 = DateTime::Fixed360Day.new(2002, 3, 5, 9, 35, 12.5, 9.quo(24))
    
    is_asserted_by { d2.difference_in_months(d1) == 1.quo(2) }
    is_asserted_by { d3.difference_in_months(d1) == 1.quo(2) }
    is_asserted_by { d4.difference_in_months(d1) == 1 }
    is_asserted_by { d5.difference_in_months(d1) == 12 }
    
  end

end

